package geniusweb.protocol.tournament.allpermutations;

import java.util.Collections;
import java.util.List;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;

import geniusweb.protocol.session.SessionSettings;
import geniusweb.protocol.tournament.TournamentProtocol;
import geniusweb.protocol.tournament.TournamentSettings;
import geniusweb.references.PartyRef;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import tudelft.utilities.immutablelist.FixedList;
import tudelft.utilities.immutablelist.Function2;
import tudelft.utilities.immutablelist.ImmutableList;
import tudelft.utilities.immutablelist.MapList;
import tudelft.utilities.immutablelist.MapThreadList;
import tudelft.utilities.immutablelist.Permutations;
import tudelft.utilities.immutablelist.PermutationsOrderedWithoutReturn;
import tudelft.utilities.immutablelist.PermutationsWithReturn;
import tudelft.utilities.immutablelist.PermutationsWithoutReturn;
import tudelft.utilities.immutablelist.Tuple;
import tudelft.utilities.immutablelist.Tuples;
import tudelft.utilities.logging.Reporter;

/**
 * Takes a number of parties and profiles and makes all permutations of them
 * with the given number of parties and profiles. All profiles must be in the
 * same domain. All sessions are based on the given sessionsettings (can be any
 * session protocol). The parties for each tournament session are added to this
 * basic sessionsettings.
 *
 */
public class AllPermutationsSettings implements TournamentSettings {
	private final List<PartyRef> parties;
	private final List<ProfileRef> profiles;
	private final boolean reuseParties;
	private final int partiesPerSession;
	private final SessionSettings sessionsettings;

	/**
	 * 
	 * @param parties           a list of {@link PartyRef}s. Must contain at
	 *                          least <partiesPerSession> elements.
	 * @param reuseParties      if true, we use PermutationsWithReturn, if false
	 *                          we use PermutationsWithoutReturn to create the
	 *                          parties.
	 * @param profiles
	 * @param partiesPerSession number of parties per session, must be at least
	 *                          2.
	 * @param sesettings        The generic {@link SessionSetting}.
	 *                          {@link SessionSettings#with(PartyWithProfile)}
	 *                          will be used to add the required
	 *                          {@link PartyWithProfile}s
	 */
	@JsonCreator
	public AllPermutationsSettings(
			@JsonProperty("parties") List<PartyRef> parties,
			@JsonProperty("reuseParties") boolean reuseParties,
			@JsonProperty("profiles") List<ProfileRef> profiles,
			@JsonProperty("partiesPerSession") int partiesPerSession,
			@JsonProperty("sessionsettings") SessionSettings sesettings) {
		if (partiesPerSession < 2)
			throw new IllegalArgumentException("partiesPerSession must be >=2");
		if (parties == null || parties.size() < partiesPerSession)
			throw new IllegalArgumentException("parties must contain at least "
					+ partiesPerSession + " PartyRefs");
		if (profiles == null || profiles.size() < 2)
			throw new IllegalArgumentException("profiles must contain at least "
					+ partiesPerSession + " ProfileRefs");
		this.partiesPerSession = partiesPerSession;
		this.reuseParties = reuseParties;
		this.parties = parties;
		this.profiles = profiles;
		this.sessionsettings = sesettings;
	}

	@Override
	public Double getMaxRunTime() {
		return permutations().size().doubleValue()
				* (sessionsettings.getMaxRunTime() + 2d);
	}

	@Override
	public TournamentProtocol getProtocol(Reporter logger) {
		AllPermutationsState state = new AllPermutationsState(this,
				permutations(), Collections.emptyList());
		return new AllPermutationsProtocol(state, logger);
	}

	/**
	 * @return {@link ImmutableList} containing all {@link SessionSettings} for
	 *         all Sessions in the correct order.
	 */
	public ImmutableList<SessionSettings> permutations() {
		ImmutableList<ImmutableList<PartyWithProfile>> partylistlist;
		ImmutableList<PartyRef> partieslist = new FixedList<PartyRef>(parties);
		ImmutableList<ProfileRef> profileslist = new FixedList<ProfileRef>(
				profiles);
		partylistlist = getParticipants(partieslist, profileslist,
				partiesPerSession, reuseParties);
		return new MapList<>(partyproflist -> createSetting(partyproflist),
				partylistlist);
	}

	@Override
	public String toString() {
		return "AllPermutationsSettings[" + parties + "," + reuseParties + ","
				+ profiles + "," + partiesPerSession + "," + sessionsettings
				+ "]";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((parties == null) ? 0 : parties.hashCode());
		result = prime * result + partiesPerSession;
		result = prime * result
				+ ((profiles == null) ? 0 : profiles.hashCode());
		result = prime * result + (reuseParties ? 1231 : 1237);
		result = prime * result
				+ ((sessionsettings == null) ? 0 : sessionsettings.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AllPermutationsSettings other = (AllPermutationsSettings) obj;
		if (parties == null) {
			if (other.parties != null)
				return false;
		} else if (!parties.equals(other.parties))
			return false;
		if (partiesPerSession != other.partiesPerSession)
			return false;
		if (profiles == null) {
			if (other.profiles != null)
				return false;
		} else if (!profiles.equals(other.profiles))
			return false;
		if (reuseParties != other.reuseParties)
			return false;
		if (sessionsettings == null) {
			if (other.sessionsettings != null)
				return false;
		} else if (!sessionsettings.equals(other.sessionsettings))
			return false;
		return true;
	}

	/**
	 * 
	 * @param partyproflist a list of {@link PartyWithProfile} settings
	 * @return {@link SessionSettings}
	 */
	private SessionSettings createSetting(
			ImmutableList<PartyWithProfile> partyproflist) {
		SessionSettings settings = sessionsettings;
		for (PartyWithProfile partyprof : partyproflist) {
			settings = settings.with(partyprof);
		}
		return settings;
	}

	/**
	 * 
	 * @param parties
	 * @param profiles
	 * @param n                    number of items to draw. Assumed >= 1.
	 * @param drawPartyWithPutback if parties can be drawn multiple times
	 * @return list of sessionsets where each sessionset contains settings for a
	 *         session: a list with {@link PartyWithProfile}s. The sessionsets
	 *         are made by making all combinations of parties and profiles.
	 *         Profiles are drawn with replace.
	 */
	private ImmutableList<ImmutableList<PartyWithProfile>> getParticipants(
			ImmutableList<PartyRef> parties, ImmutableList<ProfileRef> profiles,
			int n, boolean drawPartyWithPutback) {

		Permutations<PartyRef> partiesPermutations;
		if (drawPartyWithPutback) {
			partiesPermutations = new PermutationsWithReturn<>(parties, n);
		} else {
			partiesPermutations = new PermutationsWithoutReturn<>(parties, n);
		}

		Permutations<ProfileRef> profilesPermutations = new PermutationsOrderedWithoutReturn<>(
				profiles, n);

		// each tuple contains sesion info: a list of partyref and a list of
		// profileref's.
		Tuples<ImmutableList<PartyRef>, ImmutableList<ProfileRef>> tuples = new Tuples<>(
				partiesPermutations, profilesPermutations);

		return new MapList<Tuple<ImmutableList<PartyRef>, ImmutableList<ProfileRef>>, ImmutableList<PartyWithProfile>>(
				tuple -> partyprofilelist(tuple), tuples);
	}

	/**
	 * 
	 * @param tuple a tuple with (1) ImmutableList<PartyRef> (2)
	 *              ImmutableList<ProfileRef>
	 * @return list of PartyWithProfile, each picked from the two lists in
	 *         order.
	 */
	private ImmutableList<PartyWithProfile> partyprofilelist(
			Tuple<ImmutableList<PartyRef>, ImmutableList<ProfileRef>> tuple) {
		Function2<PartyRef, ProfileRef, PartyWithProfile> //
		makeparty = new Function2<PartyRef, ProfileRef, PartyWithProfile>() {
			@Override
			public PartyWithProfile apply(PartyRef party, ProfileRef profile) {
				return new PartyWithProfile(party, profile);
			}
		};
		return new MapThreadList<PartyWithProfile, PartyRef, ProfileRef>(
				makeparty, tuple.get1(), tuple.get2());
	}

}
